/***************************************************************************
 *   Copyright (C) 2008  Antony Martin <martin(dot)antony(at)yahoo(dot)fr> *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 10 juillet 2007
   derniere modification : 16/11/2008 */

#include <SCE/SCEMinimal.h>

#include <SCE/core/SCECMatrix.h>

#include <SCE/interface/SCENode.h>

/**
 * \file SCENode.c
 * \copydoc node
 * \file SCENode.h
 * \copydoc node
 */

/**
 * \defgroup node Nodes managment
 * \ingroup interface
 * \brief Nodes and node trees managment
 */

/** @{ */

static void SCE_Node_Init (SCE_SNode *node)
{
    node->parent = NULL;
    node->child = NULL;
#if SCE_LIST_ITERATOR_NO_MALLOC
    SCE_List_InitIt (&node->iterator);
    node->it = &node->iterator;
#else
    node->it = NULL;
#endif
    SCE_List_SetData (node->it, node);
    SCE_Matrix4_Identity (node->matrix);
    SCE_Matrix4_Identity (node->final_matrix);
    node->data = NULL;
    node->f = NULL;
    node->fparam = NULL;
}

static void SCE_Node_YouDontHaveParent (void *n)
{
    SCE_SNode *node = n;
    node->parent = NULL;
}

/**
 * \brief Creates a new node
 * \returns a newly allocated SCE_SNode on success or NULL on error
 */
SCE_SNode* SCE_Node_Create (void)
{
    SCE_SNode *node = NULL;
    if (!(node = SCE_malloc (sizeof *node)))
        goto failure;
    SCE_Node_Init (node);
    node->child = SCE_List_Create (SCE_Node_YouDontHaveParent);
    if (!(node->child = SCE_List_Create (SCE_Node_YouDontHaveParent)))
        goto failure;
#if !SCE_LIST_ITERATOR_NO_MALLOC
    if (!(node->it = SCE_List_CreateIt ()))
        goto failure;
#endif
    /* each node manages its own iterator */
    SCE_List_CanDeleteIterators (node->child, SCE_FALSE);
    goto success;
failure:
    SCE_Node_Delete (node), node = NULL;
    Logger_LogSrc ();
success:
    return node;
}

/**
 * \brief Deletes a node
 * \param node a SCE_SNode
 * \note This function does not do anything with node's data.
 */
void SCE_Node_Delete (SCE_SNode *node)
{
    if (node)
    {
        SCE_Node_Detach (node);
#if !SCE_LIST_ITERATOR_NO_MALLOC
        SCE_List_DeleteIt (node->it);
#endif
        SCE_List_Delete (node->child);
        SCE_free (node);
    }
}
/**
 * \brief Deletes a node and all its children
 * \param node a SCE_SNode
 * 
 * This function deletes a node recursivly.
 * \sa SCE_Node_Delete()
 */
void SCE_Node_DeleteRecursive (SCE_SNode *node)
{
    if (node)
    {
        SCE_SListIterator *i = NULL;
        SCE_List_ForEach (i, node->child)
            SCE_Node_DeleteRecursive (SCE_List_GetData (i));
        SCE_Node_Delete (node);
    }
}


/**
 * \brief Attaches a node to another
 * \param node a node
 * \param child a node to make child of \p node
 * \returns SCE_ERROR on error, SCE_OK otherwise.
 * 
 * This function makes the \p child node child of the \p node node. If the \p
 * child node has already a parent node, it will be detached before from its
 * current parent.
 * 
 * \warning If this function fails, the node will became orphean, even if it
 *          has a parent before.
 */
void SCE_Node_Attach (SCE_SNode *node, SCE_SNode *child)
{
    SCE_Node_Detach (child); /* impossible d'avoir deux parents */
    SCE_List_Prependl (node->child, child->it);
    child->parent = node;
}

/**
 * \brief Creates a new child to a node
 * \param node a node to which add a child
 * \returns the new child of \p node of NULL on error
 * 
 * This function creates a new node and attaches it as a child of the given
 * node.
 * 
 * \see SCE_Node_Create() SCE_Node_Attach()
 */
SCE_SNode* SCE_Node_AttachNew (SCE_SNode *node)
{
    SCE_SNode *new = NULL;
    if (!(new = SCE_Node_Create ()))
    {
        Logger_LogSrc ();
        return NULL;
    }
    SCE_Node_Attach (node, new);
    return new;
}

/**
 * \brief Detaches a node from its parent
 * \param node the node to detach
 */
void SCE_Node_Detach (SCE_SNode *node)
{
    if (node->parent)
    {
        SCE_List_Remove (node->parent->child, node->it);
        node->parent = NULL;
    }
}

/**
 * \brief
 * \param node a node
 * \param child 
 * \returns SCE_ERROR on error, SCE_OK otherwise
 */
void SCE_Node_Insert (SCE_SNode *node, SCE_SNode *child)
{
    if (child->parent)
        SCE_Node_Attach (child->parent, node);
    SCE_Node_Attach (node, child);
}


/**
 * \brief Calls SCE_CMultMatrix() with the final matrix of the given node
 */
void SCE_Node_MultMatrix (SCE_SNode *node)
{
    SCE_CMultMatrix (node->final_matrix);
}
/**
 * \brief Calls SCE_CLoadMatrix() with the final matrix of the given node
 */
void SCE_Node_LoadMatrix (SCE_SNode *node)
{
    SCE_CLoadMatrix (node->final_matrix);
}

/**
 * \brief Gets the node's matrix
 * \returns the matrix of the given node
 * 
 * This function gets the matrix of the given node. This gets the node's matrix
 * without any modification of its parent nodes. If you want to get the fully
 * transformed node's matrix, use SCE_Node_GetFinalMatrix() instead of this
 * function. This function declare \p node as non-updated by calling
 * SCE_Node_IsNotUpdated().
 * 
 * \see SCE_Node_GetFinalMatrix() SCE_Node_CopyMatrix()
 */
float* SCE_Node_GetMatrix (SCE_SNode *node)
{
    return node->matrix;
}
/**
 * \brief Copies the node's matrix
 * \param node a node
 * \param m a matrix where copy the node's matrix
 * 
 * See SCE_Node_GetMatrix() for more details about what is the node's matrix.
 * 
 * \see SCE_Node_CopyFinalMatrix() SCE_Node_GetMatrix()
 */
void SCE_Node_CopyMatrix (SCE_SNode *node, SCE_TMatrix4 m)
{
    SCE_Matrix4_Copy (m, node->matrix);
}

/**
 * \brief Gets the real node's matrix
 * \param node a node
 * \returns the real node's matrix
 * 
 * This function gets the matrix of the given node with all parent node's matrix
 * transformations applied. This is the real transformation matrix for the node,
 * not only the node's one.
 * If you just want the node's transofrmation matrix, without parent
 * transformations, use SCE_Node_GetMatrix()
 * 
 * \see SCE_Node_GetMatrix() SCE_Node_CopyFinalMatrix()
 */
float* SCE_Node_GetFinalMatrix (SCE_SNode *node)
{
    return node->final_matrix;
}
/**
 * \brief Copies the real node's matrix
 * \param node a node
 * \param m a matrix where copy the real node's matrix
 * 
 * See SCE_Node_GetFinalMatrix() for more details about what is the real node's
 * matrix.
 * 
 * \see SCE_Node_GetFinalMatrix()
 */
void SCE_Node_CopyFinalMatrix (SCE_SNode *node, SCE_TMatrix4 m)
{
    SCE_Matrix4_Copy (m, node->final_matrix);
}

/**
 * \brief 
 * \param node a node
 * \param f the callback function
 * \param p user-defined parameter to pass to \p f
 */
void SCE_Node_SetCallback (SCE_SNode *node, SCE_FNodeCallback f, void *p)
{
    node->f = f;
    node->fparam = p;
}

/**
 * \internal
 * \brief Updates a node by calling his callback function and multiply his
 * matrix with the final matrix of his parent
 * \param node the node to update
 * 
 * \see SCE_Node_UpdateRecursive()
 */
static void SCE_Node_Update (SCE_SNode *node)
{
    if (node->f)
        node->f (node, node->fparam);
    if (node->parent)
    {
        SCE_Matrix4_Mul (node->parent->final_matrix, node->matrix,
                         node->final_matrix);
    }
    else
        SCE_Matrix4_Copy (node->final_matrix, node->matrix);
}

/**
 * \brief Updates a node and all his child
 * \param node the node to update recursivly
 * 
 * \see SCE_Node_Update()
 */
void SCE_Node_UpdateRecursive (SCE_SNode *node)
{
    SCE_SListIterator *i = NULL;
    SCE_Node_Update (node);
    SCE_List_ForEach (i, node->child)
        SCE_Node_UpdateRecursive (SCE_List_GetData (i));
}


/**
 * \brief Checks if a node have a parent
 * \param node a node
 * \returns SCE_TRUE if the \p node node have a parent, SCE_FALSE otherwise.
 */
int SCE_Node_HaveParent (SCE_SNode *node)
{
    return (node->parent ? SCE_TRUE : SCE_FALSE);
}

/**
 * \brief Gets the parent of a node
 * \returns the parent node of \p node
 */
SCE_SNode* SCE_Node_GetParent (SCE_SNode *node)
{
    return node->parent;
}


/**
 * \brief Sets data of a node
 * \param data the data to set to \p node
 * \returns the old data if any
 */
void* SCE_Node_SetData (SCE_SNode *node, void *data)
{
    void *old_data = node->data;
    node->data = data;
    return old_data;
}

/**
 * \brief Gets data of a node
 * \returns the data of the \p node node
 */
void* SCE_Node_GetData (SCE_SNode *node)
{
    return node->data;
}


/**
 * \brief Calls SCE_Node_MultMatrix() if \p node is non-null,
 * SCE_CLoadIdentityMatrix() otherwise
 */
void SCE_Node_Use (SCE_SNode *node)
{
    if (node)
        SCE_Node_MultMatrix (node);
    else
        SCE_CLoadIdentityMatrix ();
}

/** @} */
